@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  */
 
@@ -229,7 +229,6 @@ on_session_exited (GdmSession       *session,
                    int               exit_code,
                    GdmSimpleSlave   *slave)
 {
-        slave->priv->session_is_running = FALSE;
         g_object_set (GDM_SLAVE (slave), "session-id", NULL, NULL);
 
         g_debug ("GdmSimpleSlave: session exited with code %d\n", exit_code);
@@ -241,7 +240,6 @@ on_session_died (GdmSession       *session,
                  int               signal_number,
                  GdmSimpleSlave   *slave)
 {
-        slave->priv->session_is_running = FALSE;
         g_object_set (GDM_SLAVE (slave), "session-id", NULL, NULL);
 
         g_debug ("GdmSimpleSlave: session died with signal %d, (%s)",
@@ -756,7 +754,8 @@ create_new_session (GdmSimpleSlave  *slave)
                                                 display_device,
                                                 display_seat_id,
                                                 display_x11_authority_file,
-                                                display_is_local);
+                                                display_is_local,
+                                                NULL);
 
         g_free (display_id);
         g_free (display_name);
@@ -1217,7 +1216,7 @@ wants_autologin (GdmSimpleSlave *slave)
         /* FIXME: handle wait-for-go */
 
         gdm_slave_get_timed_login_details (GDM_SLAVE (slave), &enabled, NULL, &delay);
-        return enabled && delay > 0;
+        return enabled && delay == 0;
 }
 
 #define INITIAL_SETUP_TRIGGER_FILE LOCALSTATEDIR "/lib/gdm/run-initial-setup"
