@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  */
 
@@ -1217,10 +1217,8 @@ x11_session_is_on_seat (GdmSlave        *slave,
         /* don't try to switch to our own session */
         if (x11_display == NULL || x11_display[0] == '\0'
             || strcmp (slave->priv->display_name, x11_display) == 0) {
-                g_free (x11_display);
                 goto out;
         }
-        g_free (x11_display);
 
         reply = g_dbus_connection_call_sync (slave->priv->connection,
                                              CK_NAME,
@@ -1242,7 +1240,6 @@ x11_session_is_on_seat (GdmSlave        *slave,
         g_variant_unref (reply);
 
         if (x11_display_device == NULL || x11_display_device[0] == '\0') {
-                g_free (x11_display_device);
                 goto out;
         }
 
@@ -1365,12 +1362,11 @@ gdm_slave_get_primary_session_id_for_user_from_ck (GdmSlave   *slave,
 {
         gboolean      can_activate_sessions;
         GError       *error;
+        const char  **sessions;
+        int           i;
         char         *primary_ssid;
         uid_t         uid;
         GVariant     *reply;
-        GVariant     *array;
-        GVariantIter  iter;
-        char         *ssid;
 
         error = NULL;
         primary_ssid = NULL;
@@ -1412,7 +1408,7 @@ gdm_slave_get_primary_session_id_for_user_from_ck (GdmSlave   *slave,
                                              CK_NAME,
                                              CK_MANAGER_PATH,
                                              CK_MANAGER_INTERFACE,
-                                             "CanSessionsForUnixUser",
+                                             "GetSessionsForUnixUser",
                                              g_variant_new ("(u)", uid),
                                              G_VARIANT_TYPE ("(ao)"),
                                              G_DBUS_CALL_FLAGS_NONE,
@@ -1426,17 +1422,16 @@ gdm_slave_get_primary_session_id_for_user_from_ck (GdmSlave   *slave,
                 return NULL;
         }
 
-        array = g_variant_get_child_value (reply, 0);
-        g_variant_iter_init (&iter, array);
-        while (g_variant_iter_loop (&iter, "(&s)", &ssid)) {
-                if (x11_session_is_on_seat (slave, ssid, slave->priv->display_seat_id)) {
-                        primary_ssid = g_strdup (ssid);
+        g_variant_get_child (reply, 0, "^a&o", &sessions);
+        for (i = 0; sessions[i] != NULL; i++) {
+                if (x11_session_is_on_seat (slave, sessions[i], slave->priv->display_seat_id)) {
+                        primary_ssid = g_strdup (sessions[i]);
                         break;
                 }
         }
 
+        g_free (sessions);
         g_variant_unref (reply);
-        g_variant_unref (array);
         return primary_ssid;
 }
 #endif
